/*
 * xpf_user.h:	Data structures the user of the xpf protocol must be aware of.
 *		Uses two imported definitions: XPFMessage and XPFInformation.
 *		It also assumes that xpfMalloc/xpfFree are available for use.
 */

#ifndef	_XPF_USER_H_
#define _XPF_USER_H_

#ifndef	_XPF_DEP_H_
#include "xpf_dep.h"
#endif	_XPF_DEP_H_

/*
 * Parameters for control operations.
 */

#define XPF_SET_GRANULARITY	1	/* min=1, max=XPF_DATA_MAXSIZE */
#define XPF_GET_GRANULARITY	2
#define XPF_SET_MAXASSOCIATIONS	3	/* >0 */
#define XPF_GET_MAXASSOCIATIONS	4
#define XPF_GET_NUMASSOCIATIONS	5
#define XPF_SET_GETMSGCONTENTS	6
#define XPF_GET_GETMSGCONTENTS	7
#define XPF_SET_PUTMSGCONTENTS	8
#define XPF_GET_PUTMSGCONTENTS	9

/*
 * Return status of Lookup. One or more of the flags below can be set.
 * Be careful not to confuse TRUE/FALSE etc. with the return codes set below.
 *
 * When the flags MERGEINFO / DELAYEDMSGS && STOREMSGS are set (in compilation),
 * the deletedinfo / delayedmesssagelist can be NULL even if
 * XPF_USER_TERMINATED / XPF_USER_LAUNCHED is set (respectively). If they are
 * NOT set, then these lists will always be NULL. If XPF_USER_MATCHEDNEW
 * is set, then XPF_USER_LAUNCHED must be true (can't always assert the reverse
 * since the message might match a pattern that has no programlist in it but
 * it can result in the creation of some new processes all the same).
 *
 * Iff there is no state-change whatsoever, the return value is 0.
 */

#define XPF_USER_FAILURE	0x80000000 /* If set, don't interpret rest */
#define XPF_USER_DELAYED	0x1	/* If set, don't interpret rest */
#define XPF_USER_MATCHEDNEW	0x2	/* If 0 then matched existing pattern */
#define XPF_USER_LAUNCHED	0x4	/* Were some processes launched */
#define XPF_USER_TERMINATED	0x8	/* Were some processes terminated */

/*
 * Return values of all other interface functions. Also used internally.
 */
#ifndef TRUE
#define FALSE			0
#define TRUE			1
#endif	TRUE

/*
 * Definitions of some flags which are used to distinguish different kinds
 * of patterns. They can be "or"ed together in any combination (they are
 * interpreted independently). At present XPF_FLAG_DELAY is the only one
 * that IS interpreted.
 */
#define XPF_FLAG_DELAY		0x1	/* store msg until matching one comes */
#define XPF_FLAG_NOSPECIFIC	0x2	/* no one can swamp some of my msgs */

/*
 * A capability returned to the creator of a packet filter: it is actually
 * a pointer to an XPFState but the user should not be aware of that.
 */
typedef XPFDataUnit	*XPFObject;

#if	DELAYEDMSGS && STOREMSGS
typedef struct _XPFMessageList {
	XPFMessage	message;
	int		length;
	struct _XPFMessageList *next;	/* NULL terminated */
} XPFMessageList;
#endif	DELAYEDMSGS && STOREMSGS

#if	MERGEINFO
typedef struct _XPFInfoList {
	XPFInformation	information;
	struct _XPFInfoList *next;	/* NULL terminated */
} XPFInfoList;
#endif	MERGEINFO

typedef struct _XPFSelector {
	short		offset;
#if	COUNTUNIT
	char		countunit;
	char		level;
#else	COUNTUNIT
	short		level;
#endif	COUNTUNIT
} XPFSelector;

/*
 * The types of data that can be extracted from the incoming packet.
 * Note that these correspond to the sizes of the data types: hence,
 * XPF allows you to extract any number of bytes, AND it can perform
 * optimizations for well-known types, so long as the number of bytes
 * selected is <= XPF_DATA_MAXSIZE, currently set at sizeof(unsigned long).
 * Note that if you specify type=4, 4 bytes are accessed from consecutive
 * locations: which is not the same as "int" since int-pointers HAVE to
 * start at word boundaries whereas 4 bytes can be accessed from any
 * location. XPF accesses 4 consecutive bytes if the type is "XPF_DATA_INT"
 * and the offset does not start at a long-word boundary. XPF always
 * access 2 consecutive bytes if the type is "XPF_DATA_SHORT" since the
 * above "checks" are probably not worth the effort.
 *
 * Note that if nothing is selected (the value selected is NULL), then
 * in effect, the extractor matches ANY INCOMING PACKET and the value
 * obtained from the packet will be an array of XPF_DATA_MAXSIZE 0s.
 */

#define XPF_DATA_NONE	0			/* always matches, extracts nothing  */
#define XPF_DATA_CHAR	1
#define XPF_DATA_SHORT	(XPF_DATA_MAXSIZE + 1)
#define XPF_DATA_INT	(XPF_DATA_MAXSIZE + 2)
#if	STRINGTYPE && !WORDADDRESS
#define XPF_DATA_STRING	(XPF_DATA_MAXSIZE + 3)
#endif	STRINGTYPE && !WORDADDRESS

typedef struct _XPFExtractor {
	int		datatype;		/* char enough but C pads anyway */
	XPFSelector	selector;
	XPFDataUnit	mask[XPF_DATA_MAXSIZE];
} XPFExtractor;

#if	STRINGTYPE && !WORDADDRESS
#define xpfSizeFromType(type) \
((type == XPF_DATA_NONE) ? 0 :\
    ((type == XPF_DATA_INT) ? 4 :\
	((type == XPF_DATA_SHORT) ? 2 :\
	    ((type == XPF_DATA_CHAR) ? 1 :\
		((type == XPF_DATA_STRING) ? XPF_DATA_MAXSIZE:\
		    (((type <= XPF_DATA_MAXSIZE) && (type >= 0)) ? type : XPF_DATA_MAXSIZE)\
		)\
	    )\
	)\
    )\
)
#else	STRINGTYPE && !WORDADDRESS
#define xpfSizeFromType(type) \
((type == XPF_DATA_NONE) ? 0 :\
    ((type == XPF_DATA_INT) ? 4 :\
	((type == XPF_DATA_SHORT) ? 2 :\
	    ((type == XPF_DATA_CHAR) ? 1 :\
		(((type <= XPF_DATA_MAXSIZE) && (type >= 0)) ? type : XPF_DATA_MAXSIZE)\
	    )\
	)\
    )\
)
#endif	STRINGTYPE && !WORDADDRESS

/*
 * The normal-case: headertype & 0x3 == 0 => the header is of a fixed size.
 * Other combinations have obvious meanings: they tell us how to get to the
 * beginning of the header of the next protocol, which is simply the beginning
 * of the header of the current protocol + size of the header computed above.
 *
 * The beginning of the header of the protocol immediately above XPF is
 * considered to be the first-byte of the packet, i.e., offset 0 (e.g., ETH).
 * Note that for header-size computations, only type=CHAR, SHORT, INT are used.
 */

/* What's CURRENT? see segments below. These can be |ed to-gether */
#define XPF_HEADER_FIXED	0x0	/* new protocol has fixed size hdr */
#if	VARHEADER
#define XPF_HEADER_VARIABLE	0x1	/* new protocol has variable size hdr */
#endif	VARHEADER
#define XPF_HEADER_CURRENT	0x2	/* Have'nt crossed protl bdry yet */

/*
 * Note: in SIPP, even if the main header is of variable length, the length is known to the fellow who is sending the packet. So, that person
 * can give filters that have offsets beginning from the main header rather than the current header. Otherwise, we need negative offsets from
 * the supplementary headers to access the main header and that complicates the implementation a lot. But this is a pain for the user: can't
 * we do something like two kinds of offsets: relative to current header and absolute offsets? Relative is the default, absolute is also needed.
 * But we must know which header is the main header for a particular supplementary header --> maintain a separate array for that. But suppose
 * we need relative offsets w.r.t. each previous header, then we surely need a level number to access it: level 0 is the default = current.
 * level 1... etc refer to headers before this one. Put that as a part of the extractor. We don't need supplementary header for that once we
 * add level# to a extractor, i.e., give a "peek" facility.
 */

typedef int XPFFixedHeader;

typedef struct _XPFVariableHeader {
	int		multiplier;	/* +ve => multiply, -ve => divide */
	XPFExtractor	extractor;
} XPFVariableHeader;

typedef struct _XPFHeader {
	int	headertype;	/* C rounds structures to integers anyway */
	union {
#if	VARHEADER
		XPFVariableHeader	*variable;
#endif	VARHEADER
		XPFFixedHeader		fixed;
	} size;
} XPFHeader;

/*
 * Programs just have an extractor. They are used to extract
 * information from the incoming-packet to add elements to a filter.
 * Filters also have a value with which the extracted information
 * from the packet is compared for a match.
 */

typedef struct _XPFProgram {
	struct _XPFProgram *next;
	XPFExtractor	extractor;
	int		refcount;
} XPFProgram;

/*
 * XPF does not check if the value is a subset of the mask in the extractor:
 * the user should do it. A user can use this to setup a dummy-filter that
 * matches nothing if reqd.: there is no provision for an ALWAYS-FAIL filter.
 */
typedef struct _XPFFilter {
	struct _XPFFilter *next;	/* next filter in this segment */
	XPFExtractor	extractor;
	XPFDataUnit	value[XPF_DATA_MAXSIZE];
	int		refcount;
} XPFFilter;

#if	COUNTUNIT
#define XPF_USER_GETCOUNTUNIT(filter)	* (filter)->extractor.selector.countunit
#else	COUNTUNIT
#define XPF_USER_GETCOUNTUNIT(filter)
#endif	COUNTUNIT

/*
 * A segment is a portion of the pattern with the same header which can
 * be treated as a linear sequence of filters = (mask, value) pairs.
 * It is used for branching in the decision DAG (see xpf_internal.h).
 * The decision point can be in the middle of the same protocol or in
 * a new protocol.
 */

typedef struct _XPFPattern {
	XPFHeader	header;		/* size of the hdr of the next seg. */
	XPFFilter	*filterlist;	/* list of filters for this segment */
	XPFProgram	*programlist;	/* list of programs  for this seg. */
	int		flags;		/* XPF_FLAG_DELAY|XPF_FLAG_NOSPECIFIC */
	short		alarm;		/* max #of concurrent processes */
	short		processnum;	/* not used by user */
	struct _XPFPattern *next;	/* not used directly: points to pattern of next protocol */
} XPFPattern;

/*
 * These are some contructor macros: use only when the memory
 * for the structure being created has to be allocated.
 */

#define xpfCreatePattern(p) p = (XPFPattern *)xpfMalloc(sizeof(XPFPattern));

#define xpfCreateFilter(f) f = (XPFFilter *)xpfMalloc(sizeof(XPFFilter));

#define xpfCreateProgram(r) r = (XPFProgram *)xpfMalloc(sizeof(XPFProgram));

#define XPFCreateVariableHeader(v) v = (XPFVariableHeader *)xpfMalloc(sizeof(XPFVariableHeader));

#if	MERGEINFO
#define xpfCreateInfoListElement(i) i = (XPFInfoList *)xpfMalloc(sizeof(XPFInfoList));
#endif	MERGEINFO

#if	DELAYEDMSGS && STOREMSGS
#define xpfCreateMessageListElement(m) m = (XPFMessageList *)xpfMalloc(sizeof(XPFMessageList));
#endif	DELAYEDMSGS && STOREMSGS

/*
 * These are some initialization macros: should be used after
 * allocating memory (if reqd) and initializing all the "sub-components",
 * i.e., xpfInitializeExtractor(e) should be called AFTER calling
 * xpfInitilizeSelector(&(e)->selector, offset, countunit).
 */

#if	COUNTUNIT
#define xpfInitializeSelector(_s, _offset, _level, _countunit)\
{\
	(_s)->offset = (_offset);\
	(_s)->level = (_level);\
	(_s)->countunit = (_countunit);\
}
#else	COUNTUNIT
#define xpfInitializeSelector(_s, _offset, _level)\
{\
	(_s)->offset = (_offset);\
	(_s)->level = (_level);\
}
#endif	COUNTUNIT

#if	STRINGTYPE && !WORDADDRESS
#define xpfInitializeExtractor(_e, _datatype, _mask) \
{\
	(_e)->datatype = (_datatype);\
	*(unsigned long *)(_e)->mask = 0;	/* zero out unused part */\
	switch ((_datatype)) \
	{\
	case XPF_DATA_CHAR:\
		*(unsigned char *)(_e)->mask = (unsigned char)(_mask); break;\
	case XPF_DATA_SHORT:\
		*(unsigned short *)(_e)->mask = (unsigned short)(_mask); break;\
	case XPF_DATA_INT:\
		*(unsigned int *)(_e)->mask = (unsigned int)(_mask); break;\
	case XPF_DATA_STRING:\
		xpfStrncpy((_e)->mask, (char *)(_mask), XPF_DATA_MAXSIZE); break;\
	default:\
		xpfMemcpy((char *)(_e)->mask, (char *)(_mask), (_datatype)); break;\
	}\
}
#else	STRINGTYPE && !WORDADDRESS
#define xpfInitializeExtractor(_e, _datatype, _mask) \
{\
	(_e)->datatype = (_datatype);\
	*(unsigned long *)(_e)->mask = 0;	/* zero out unused part */\
	switch ((_datatype)) \
	{\
	case XPF_DATA_CHAR:\
		*(unsigned char *)(_e)->mask = (unsigned char)(_mask); break;\
	case XPF_DATA_SHORT:\
		*(unsigned short *)(_e)->mask = (unsigned short)(_mask); break;\
	case XPF_DATA_INT:\
		*(unsigned int *)(_e)->mask = (unsigned int)(_mask); break;\
	default:\
		xpfMemcpy((char *)(_e)->mask, (char *)(_mask), (_datatype)); break;\
	}\
}
#endif	STRINGTYPE && !WORDADDRESS

#if	VARHEADER
#if	COUNTUNIT
#define xpfInitializeVariableHeader(_h, _varpart, _offset, _countunit, _datatype, _mask, _multiplier) \
{\
	(_h)->headertype = XPF_HEADER_VARIABLE;\
	(_h)->size.variable = _varpart;\
	xpfInitializeSelector(&(_h)->size.variable->extractor.selector, _offset, 0, _countunit);\
	xpfInitializeExtractor(&(_h)->size.variable->extractor, _datatype, _mask);\
	(_h)->size.variable->multiplier =_multiplier;\
}
#else	COUNTUNIT
#define xpfInitializeVariableHeader(_h, _varpart, _offset, _datatype, _mask, _multiplier) \
{\
	(_h)->headertype = XPF_HEADER_VARIABLE;\
	(_h)->size.variable = _varpart;\
	xpfInitializeSelector(&(_h)->size.variable->extractor.selector, _offset, 0);\
	xpfInitializeExtractor(&(_h)->size.variable->extractor, _datatype, _mask);\
	(_h)->size.variable->multiplier =_multiplier;\
}
#endif	COUNTUNIT
#endif	VARHEADER

#define xpfInitializeFixedHeader(_h, _size) \
{\
	(_h)->headertype = XPF_HEADER_FIXED;\
	(_h)->size.fixed = _size;\
}

#if	STRINGTYPE && !WORDADDRESS
#define xpfInitializeFilter(_f, _value, _next) \
{\
	(_f)->refcount = 1;\
	(_f)->next = (_next);\
	*(unsigned long *)(_f)->value = 0;	/* zero out unused part */\
	switch ((_f)->extractor.datatype) \
	{\
	case XPF_DATA_CHAR:\
		*(unsigned char *)(_f)->value = (unsigned char)(_value); break;\
	case XPF_DATA_SHORT:\
		*(unsigned short *)(_f)->value = (unsigned short)(_value); break;\
	case XPF_DATA_INT:\
		*(unsigned int *)(_f)->value = (unsigned int)(_value); break;\
	case XPF_DATA_STRING:\
		xpfStrncpy((_f)->value, (char *)(_value), XPF_DATA_MAXSIZE); break;\
	default:\
		xpfMemcpy((char *)(_f)->value, (char *)(_value), (_f)->extractor.datatype); break;\
	}\
}
#else	STRINGTYPE && !WORDADDRESS
#define xpfInitializeFilter(_f, _value, _next) \
{\
	(_f)->refcount = 1;\
	(_f)->next = (_next);\
	*(unsigned long *)(_f)->value = 0;	/* zero out unused part */\
	switch ((_f)->extractor.datatype) \
	{\
	case XPF_DATA_CHAR:\
		*(unsigned char *)(_f)->value = (unsigned char)(_value); break;\
	case XPF_DATA_SHORT:\
		*(unsigned short *)(_f)->value = (unsigned short)(_value); break;\
	case XPF_DATA_INT:\
		*(unsigned int *)(_f)->value = (unsigned int)(_value); break;\
	default:\
		xpfMemcpy((char *)(_f)->value, (char *)(_value), (_f)->extractor.datatype); break;\
	}\
}
#endif	STRINGTYPE && !WORDADDRESS

#define xpfInitializeProgram(_r, _next) \
{\
	(_r)->refcount = 1;\
	(_r)->next = _next;\
}

#define xpfInitializePattern(_p, _filterlist, _programlist, _flags, _alarm, _next) \
{\
	(_p)->filterlist = _filterlist;\
	(_p)->programlist = _programlist;\
	(_p)->flags = _flags;\
	(_p)->alarm = _alarm;\
	(_p)->next = _next;\
}

/* These two are probably overkill! */

#if	MERGEINFO
#define xpfInitializeInfoListElement(i, info, next) \
{\
	(i)->information = info;\
	(i)->next = next;\
}
#endif	MERGEINFO

#if	DELAYEDMSGS && STOREMSGS
#define xpfInitializeMessageListElement(m, msg, len, next) \
{\
	(m)->message = msg;\
	(m)->length = len;\
	(m)->next = next;\
}
#endif	DELAYEDMSGS && STOREMSGS

/*
 * These are some destructor macros. Not really necessary!
 */

#define xpfDestroySelector(s) xpfFree(s, sizeof(XPFSelector))

#define xpfDestroyExtractor(e) xpfFree(e, sizeof(XPFExtractor))

#define xpfDestroyFilter(f) xpfFree(f, sizeof(XPFFilter))

#define xpfDestroyProgram(r) xpfFree(r, sizeof(XPFProgram))

#define xpfDestroyPattern(p) xpfFree(p, sizeof(XPFPattern))

/*
 * Completely ruins the given filterlist. DOES NOT MAKE IT NULL.
 */
#define xpfDestroyFilterList(statlist)\
{\
	XPFFilter	*_s, *_ts;\
\
	_s = statlist;\
	while (_s != NULL) {\
		_ts = _s;\
		_s = _s->next;\
		xpfFree((char *)_ts, sizeof(XPFFilter));\
	}\
	/* statlist = NULL; */\
}

/*
 * Completely ruins the given programlist. DOES NOT MAKE IT NULL.
 */
#define xpfDestroyProgramList(dynlist)\
{\
	XPFProgram	*_d, *_td;\
\
	_d = dynlist;\
	while (_d != NULL) {\
		_td = _d;\
		_d = _d->next;\
		xpfFree((char *)_td, sizeof(XPFProgram));\
	}\
	/* dynlist = NULL; */\
}

#if	VARHEADER
#define xpfDestroyVariableHeader(h) xpfFree((char *)(h)->size.variable, sizeof(XPFVariableHeader));
#endif	VARHEADER

#if	VARHEADER
/*
 * Completely ruins the given patternlist. DOES NOT MAKE IT NULL.
 */
#define xpfDestroyPatternList(patternlist)\
{\
	XPFPattern	*_k, *_tk;\
\
	_k = patternlist;\
	while (_k != NULL) {\
		xpfDestroyFilterList(_k->filterlist);\
		if (_k->programlist != NULL) xpfDestroyProgramList(_k->programlist);\
		_tk = _k;\
		_k = _k->next;\
		if (_tk->header.headertype & XPF_HEADER_VARIABLE) xpfDestroyVariableHeader(&_tk->header);\
		xpfFree((char *)_tk, sizeof(XPFPattern));\
	}\
	/* patternlist = NULL; */\
}
#else	VARHEADER
/*
 * Completely ruins the given patternlist.
 */
#define xpfDestroyPatternList(patternlist)\
{\
	XPFPattern	*_k, *_tk;\
\
	_k = patternlist;\
	while (_k != NULL) {\
		xpfDestroyFilterList(_k->filterlist);\
		xpfDestroyProgramList(_k->programlist);\
		_tk = _k;\
		_k = _k->next;\
		xpfFree((char *)_tk, sizeof(XPFPattern));\
	}\
}
#endif	VARHEADER

#define xpfDestroyHeader(h) xpfFree((char *)(h), sizeof(XPFHeader));

#if	MERGEINFO
#define xpfDestroyInfoListElement(i) xpfFree((char *)i, sizeof(XPFInfoList));
#endif	MERGEINFO

#if	DELAYEDMSGS && STOREMSGS
#define xpfDestroyMessageListElement(m) xpfFree((char *)m, sizeof(XPFMessageList));
#endif	DELAYEDMSGS && STOREMSGS

/*
 * These are some simple debug-o/p macros: they DON'T print '\n' at the end.
 * We don't really need routines to print out a selector/extractor.
 * All values/masks are printed in "full" as integers using %x everywhere.
 */

#if	COUNTUNIT
#define xpfPrintFilterList(_filterlist) \
{\
	XPFFilter	*_s;\
\
	xpfPrintf("[");\
	_s = _filterlist;\
	while (_s != NULL) {\
		xpfPrintf("%x@%x,%x=%x ", xpfSizeFromType((_s)->extractor.datatype), (_s)->extractor.selector.countunit * (_s)->extractor.selector.offset, *(unsigned long *)((_s)->extractor.mask), *(unsigned long *)((_s)->value));\
		(_s) = (_s)->next;\
	}\
	xpfPrintf("] ");\
}
#else	COUNTUNIT
#define xpfPrintFilterList(_filterlist) \
{\
	XPFFilter	*_s;\
\
	xpfPrintf("[");\
	_s = _filterlist;\
	while (_s != NULL) {\
		xpfPrintf("%x@%x,%x=%x ", xpfSizeFromType((_s)->extractor.datatype), (_s)->extractor.selector.offset, *(unsigned long *)((_s)->extractor.mask), *(unsigned long *)((_s)->value));\
		(_s) = (_s)->next;\
	}\
	xpfPrintf("] ");\
}
#endif	COUNTUNIT

#if	COUNTUNIT
#define xpfPrintProgramList(_programlist) \
{\
	XPFProgram	*_d;\
\
	_d = _programlist;\
	xpfPrintf("(");\
	while (_d != NULL) {\
		xpfPrintf("%x@%x,%x ", xpfSizeFromType((_d)->extractor.datatype), (_d)->extractor.selector.countunit * (_d)->extractor.selector.offset, *(unsigned long *)((_d)->extractor.mask));\
		(_d) = (_d)->next;\
	}\
	xpfPrintf(") ");\
}
#else	COUNTUNIT
#define xpfPrintProgramList(_programlist) \
{\
	XPFProgram	*_d;\
\
	_d = _programlist;\
	xpfPrintf("(");\
	while (_d != NULL) {\
		xpfPrintf("%x@%x,%x ", xpfSizeFromType((_d)->extractor.datatype), (_d)->extractor.selector.offset, *(unsigned long *)((_d)->extractor.mask));\
		(_d) = (_d)->next;\
	}\
	xpfPrintf(") ");\
}
#endif	COUNTUNIT

#if	VARHEADER
#if	COUNTUNIT
#define xpfPrintHeader(_h) { \
	if ((_h)->headertype & XPF_HEADER_CURRENT) xpfPrintf("C");\
	else xpfPrintf("N");\
	if ((_h)->headertype & XPF_HEADER_VARIABLE) {\
		xpfPrintf("V:%x@%x,%x*%x ", xpfSizeFromType((_h)->size.variable->extractor.datatype), (_h)->size.variable->extractor.selector.countunit * (_h)->size.variable->extractor.selector.offset, *(unsigned long *)(_h)->size.variable->extractor.mask, (_h)->size.variable->multiplier);\
	}\
	else {\
		xpfPrintf("F:%x ", (_h)->size.fixed);\
	}\
}
#else	COUNTUNIT
#define xpfPrintHeader(_h) { \
	if ((_h)->headertype & XPF_HEADER_CURRENT) xpfPrintf("C");\
	else xpfPrintf("N");\
	if ((_h)->headertype & XPF_HEADER_VARIABLE) {\
		xpfPrintf("V:%x@%x,%x*%x ", xpfSizeFromType((_h)->size.variable->extractor.datatype), (_h)->size.variable->extractor.selector.offset, *(unsigned long *)(_h)->size.variable->extractor.mask, (_h)->size.variable->multiplier);\
	}\
	else {\
		xpfPrintf("F:%x ", (_h)->size.fixed);\
	}\
}
#endif	COUNTUNIT
#else	VARHEADER
#define xpfPrintHeader(_h) { \
	if ((_h)->headertype & XPF_HEADER_CURRENT) xpfPrintf("C");\
	else xpfPrintf("N");\
	xpfPrintf("F:%x ", (_h)->size.fixed);\
}
#endif	VARHEADER

#define xpfPrintPatternList(_patternlist) \
{\
	XPFPattern	*_k;\
\
	_k = _patternlist;\
	while (_k != NULL) {\
		xpfPrintf("{");\
		xpfPrintHeader(&(_k)->header);\
		xpfPrintFilterList((_k)->filterlist);\
		xpfPrintProgramList((_k)->programlist);\
		_k = _k->next;\
		xpfPrintf("} ");\
	}\
}

#endif	_XPF_USER_H_
